import socket
import struct
import socketserver
from crypt import crypt
from ssl import SSLContext, SSLSocket

from sqlalchemy import create_engine, Column, Integer, String, Table, ForeignKey
from sqlalchemy.orm import sessionmaker, relationship, aliased
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property

from ldapserver import SimpleLDAPRequestHandler, sasl
from ldapserver.ldap import SearchScope
from ldapserver_sqlalchemy import SQLDirectory

Base = declarative_base()

engine = create_engine('sqlite:///db.sqlite', echo=True)
Session = sessionmaker(bind=engine)
session = Session()

user_groups = Table('user_groups', Base.metadata,
	Column('user_id', ForeignKey('users.id'), primary_key=True),
	Column('group_id', ForeignKey('groups.id'), primary_key=True)
)

class User(Base):
	__tablename__ = 'users'

	id = Column(Integer, primary_key=True)
	loginname = Column(String, unique=True, nullable=False)
	displayname = Column(String, nullable=False, default='')
	email = Column(String)
	pwhash = Column(String)
	groups = relationship('Group', secondary=user_groups, back_populates='users')

	ldap_gid = 1
	ldap_sn = ' '

	@hybrid_property
	def homedirectory(self):
		return '/home/' + self.loginname

	# Write-only property
	def password(self, value):
		self.pwhash = crypt(value)
	password = property(fset=password)

	def check_password(self, password):
		return self.pwhash is not None and crypt(password, self.pwhash) == self.pwhash

class Group(Base):
	__tablename__ = 'groups'

	id = Column(Integer, primary_key=True)
	name = Column(String, unique=True, nullable=False)
	description = Column(String, nullable=False, default='')
	users = relationship('User', secondary=user_groups, back_populates='groups')

Base.metadata.create_all(engine)

sqldirectory = SQLDirectory(session, Base)
sqldirectory.register_model(
	model=User,
	attribute_map={
		'cn': 'displayname',
		'displayname': 'displayname',
		'gidNumber': 'ldap_gid',
		'givenname': 'displayname',
		'homeDirectory': 'homedirectory',
		'mail': 'email',
		'sn': 'ldap_sn',
		'uid': 'loginname',
		'uidNumber': 'id',
		'memberOf': 'groups',
	},
	objectclasses=[b'top', b'inetOrgPerson', b'organizationalPerson', b'person', b'posixAccount'],
	rdn_attr='uid',
	dn_base='ou=users,dc=example,dc=com'
)

sqldirectory.register_model(
	model=Group,
	attribute_map={
		'cn': 'name',
		'description': 'description',
		'gidNumber': 'id',
		'uniqueMember': 'users',
	},
	objectclasses=[b'top', b'posixGroup', b'groupOfUniqueNames'],
	rdn_attr='cn',
	dn_base='ou=groups,dc=example,dc=com'
)

ssl_context = SSLContext()
ssl_context.load_cert_chain('devcert.crt', 'devcert.key')

class RequestHandler(SimpleLDAPRequestHandler):
	ssl_context = ssl_context

	supports_sasl_digest_md5 = True

	def do_bind_sasl_digest_md5(self, username, realm, host, serv_name=None, authzid=None, charset='utf-8'):
		return [(None, sasl.digest_md5.credential_digest(username, realm, pw)) for pw in ['foobar', 'abcdef']]

	supports_sasl_plain = True

	def do_bind_sasl_plain(self, identity, password, authzid=None):
		if not isinstance(self.request, SSLSocket):
			raise LDAPConfidentialityRequired()
		if authzid is not None and authzid != identity:
			raise LDAPProtocolError()
		if identity.startswith('dn:'):
			return self.do_bind_simple_authenticated(identity[3:], password.encode())
		user = session.query(User).filter_by(loginname=identity).one_or_none()
		if user is None or not user.check_password(password):
			raise LDAPInvalidCredentials()
		return user

	def do_bind_simple_authenticated(self, dn, password):
		if not isinstance(self.request, SSLSocket):
			raise LDAPConfidentialityRequired()
		try:
			password = password.decode()
		except UnicodeDecodeError:
			raise LDAPInvalidCredentials()
		try:
			user = session.query(User).filter(sqldirectory.models[User].filter_dn(dn, SearchScope.baseObject)).one_or_none()
		except ValueError:
			raise LDAPInvalidCredentials()
		if user is None or not user.check_password(password):
			raise LDAPInvalidCredentials()
		return user

	@property
	def supports_sasl_external(self):
		return self.request.family == socket.AF_UNIX

	def do_bind_sasl_external(self, authzid=None):
		pid, uid, gid = sasl.external.get_peercred(self.request)
		if not authzid:
			return None
		if uid != 0:
			raise LDAPInvalidCredentials()
		user = session.query(User).filter_by(loginname=authzid).one_or_none()
		if user is None:
			raise LDAPNoSuchObject()
		return user

	supports_whoami = True

	def do_whoami(self):
		if self.bind_object is None:
			return ''
		return self.bind_object.loginname

	def do_search(self, baseobj, scope, filter):
		yield from super().do_search(baseobj, scope, filter)
		if self.bind_object is not None:
			yield from sqldirectory.search(baseobj, scope, filter)

socketserver.ForkingTCPServer(('127.0.0.1', 1337), RequestHandler).serve_forever()
#socketserver.UnixStreamServer('/tmp/ldapd.sock', RequestHandler).serve_forever()
