# Decoding code based on python-ldap3

def compute_ldap_message_size(data): # from ldap3
	ret_value = -1
	if len(data) > 2:
		if data[1] <= 127:  # BER definite length - short form. Highest bit of byte 1 is 0, message length is in the last 7 bits - Value can be up to 127 bytes long
			ret_value = data[1] + 2
		else:  # BER definite length - long form. Highest bit of byte 1 is 1, last 7 bits counts the number of following octets containing the value length
			bytes_length = data[1] - 128
			if len(data) >= bytes_length + 2:
				value_length = 0
				cont = bytes_length
				for byte in data[2:2 + bytes_length]:
					cont -= 1
					value_length += byte * (256 ** cont)
				ret_value = value_length + 2 + bytes_length
	return ret_value

CLASSES = {(False, False): 0,  # Universal
		   (False, True): 1,  # Application
		   (True, False): 2,  # Context
		   (True, True): 3}  # Private

# a fast BER decoder for LDAP responses only
def compute_ber_size(data):
	"""
	Compute size according to BER definite length rules
	Returns size of value and value offset
	"""

	if data[1] <= 127:  # BER definite length - short form. Highest bit of byte 1 is 0, message length is in the last 7 bits - Value can be up to 127 bytes long
		return data[1], 2
	else:  # BER definite length - long form. Highest bit of byte 1 is 1, last 7 bits counts the number of following octets containing the value length
		bytes_length = data[1] - 128
		value_length = 0
		cont = bytes_length
		for byte in data[2: 2 + bytes_length]:
			cont -= 1
			value_length += byte * (256 ** cont)
		return value_length, bytes_length + 2


def decode_message_fast(message):
	ber_len, ber_value_offset = compute_ber_size(get_bytes(message[:10]))  # get start of sequence, at maximum 3 bytes for length
	decoded = decode_sequence(message, ber_value_offset, ber_len + ber_value_offset, LDAP_MESSAGE_CONTEXT)
	return {
		'messageID': decoded[0][3],
		'protocolOp': decoded[1][2],
		'payload': decoded[1][3],
		'controls': decoded[2][3] if len(decoded) == 3 else None
	}

def decode_sequence(message, start, stop, context_decoders=None):
	decoded = []
	while start < stop:
		octet = get_byte(message[start])
		ber_class = CLASSES[(bool(octet & 0b10000000), bool(octet & 0b01000000))]
		ber_constructed = bool(octet & 0b00100000)
		ber_type = octet & 0b00011111
		ber_decoder = DECODERS[(ber_class, octet & 0b00011111)] if ber_class < 2 else None
		ber_len, ber_value_offset = compute_ber_size(get_bytes(message[start: start + 10]))
		start += ber_value_offset
		if ber_decoder:
			value = ber_decoder(message, start, start + ber_len, context_decoders)  # call value decode function
		else:
			# try:
			value = context_decoders[ber_type](message, start, start + ber_len)  # call value decode function for context class
			# except KeyError:
			#	 if ber_type == 3:  # Referral in result
			#		 value = decode_sequence(message, start, start + ber_len)
			#	 else:
			#		 raise  # re-raise, should never happen
		decoded.append((ber_class, ber_constructed, ber_type, value))
		start += ber_len

	return decoded


def decode_integer(message, start, stop, context_decoders=None):
	first = message[start]
	value = -1 if get_byte(first) & 0x80 else 0
	for octet in message[start: stop]:
		value = value << 8 | get_byte(octet)

	return value


def decode_octet_string(message, start, stop, context_decoders=None):
	return message[start: stop]


def decode_boolean(message, start, stop, context_decoders=None):
	return False if message[start: stop] == 0 else True

def decode_search_request(message, start, stop, context_decoders=None):
	decoded = decode_sequence(message, start, stop, SEARCH_FILTER_CONTEXT)
	return {
		'baseObject': decoded[0][3],
		'scope': decoded[1][3],
		'filter': decoded[6][3],
		'attributes': [item[3] for item in decoded[7][3]],
	}

def decode_bind_request(message, start, stop, context_decoders=None):
	decoded = decode_sequence(message, start, stop, BIND_REQUEST_CONTEXT)
	return {
		'version': decoded[0][3],
		'name': decoded[1][3],
		'authentication': decoded[2][3],
	}

def decode_filter_and(message, start, stop, context_decoders=None):
	decoded = decode_sequence(message, start, stop, SEARCH_FILTER_CONTEXT)
	return ('and',) + tuple(item[3] for item in decoded)

def decode_filter_or(message, start, stop, context_decoders=None):
	decoded = decode_sequence(message, start, stop, SEARCH_FILTER_CONTEXT)
	return ('or',) + tuple(item[3] for item in decoded)

def decode_filter_not(message, start, stop, context_decoders=None):
	decoded = decode_sequence(message, start, stop, SEARCH_FILTER_CONTEXT)
	return ('not', decoded[0][3])

def decode_filter_equal(message, start, stop, context_decoders=None):
	decoded = decode_sequence(message, start, stop, SEARCH_FILTER_CONTEXT)
	return ('equal', decoded[0][3], decoded[1][3])

def decode_filter_present(message, start, stop, context_decoders=None):
	#decoded = decode_sequence(message, start, stop, SEARCH_FILTER_CONTEXT)
	#return ('present', ecoded[0][3])
	return ('present', decode_octet_string(message, start, stop))

def decode_filter_unknown(message, start, stop, context_decoders=None):
	return ('unknown',)

def decode_controls(message, start, stop, context_decoders=None):
	return decode_sequence(message, start, stop, CONTROLS_CONTEXT)

######

if str is not bytes:  # Python 3
	def get_byte(x):
		return x

	def get_bytes(x):
		return x
else:  # Python 2
	def get_byte(x):
		return ord(x)

	def get_bytes(x):
		return bytearray(x)

DECODERS = {
	# Universal
	(0, 1): decode_boolean,  # Boolean
	(0, 2): decode_integer,  # Integer
	(0, 4): decode_octet_string,  # Octet String
	(0, 10): decode_integer,  # Enumerated
	(0, 16): decode_sequence,  # Sequence
	(0, 17): decode_sequence,  # Set
	# Application
	(1, 0): decode_bind_request,   # Bind
	(1, 2): decode_sequence,       # Unbind
	(1, 3): decode_search_request, # Search request
	(1, 6): decode_octet_string,   # Modify request
	(1, 8): decode_octet_string,   # Add request
	(1, 10): decode_octet_string,  # Del request
	(1, 12): decode_octet_string,  # ModDN request
	(1, 14): decode_octet_string,  # Compare request
	(1, 16): decode_sequence,      # Abandon
	(1, 23): decode_octet_string,  # Extended Request
}

BIND_REQUEST_CONTEXT = {
	0: decode_octet_string, # simple
	3: decode_octet_string, # SaslCredentials
}

SEARCH_FILTER_CONTEXT = {
	0: decode_filter_and,      # and
	1: decode_filter_or,       # or
	2: decode_filter_not,      # not
	3: decode_filter_equal,    # equalityMatch
	4: decode_filter_unknown,  # substrings
	5: decode_filter_unknown,  # greaterOrEqual
	6: decode_filter_unknown,  # lessOrEqual
	7: decode_filter_present,  # present
	8: decode_filter_unknown,  # approxMatch
	9: decode_filter_unknown,  # extensibleMatch
}

LDAP_MESSAGE_CONTEXT = {
	0: decode_controls,  # Controls
	3: decode_sequence  # Referral
}

CONTROLS_CONTEXT = {
	0: decode_sequence  # Control
}
