from ldapserver.directory import BaseDirectory, FilterMixin
from ldapserver.util import encode_attribute, CaseInsensitiveDict
from ldapserver.dn import DN, RDN
from ldapserver.ldap import SearchScope, FilterAnd, FilterOr, FilterNot, FilterEqual, FilterPresent

from sqlalchemy import or_, and_, Integer, String
from sqlalchemy.orm import RelationshipProperty

class SQLModelDirectory(FilterMixin, BaseDirectory):
	def __init__(self, directory, model, attribute_map=None, objectclasses=None, rdn_attr='uid', dn_base=''):
		self.directory = directory
		self.model = model
		self.attribute_map = CaseInsensitiveDict(attribute_map or {})
		self.objectclasses = []
		for value in (objectclasses or []):
			value = value.lower()
			if isinstance(value, str):
				value = value.encode()
			self.objectclasses.append(value)
		self.rdn_attr = rdn_attr
		self.dn_base = DN(dn_base)

	def search_fetch(self, result):
		if result is False:
			return []
		elif result is True:
			objs = self.directory.session.query(self.model)
		else:
			objs = self.directory.session.query(self.model).filter(result)
		for obj in objs:
			attrs = {}
			for ldap_name, attr_name in self.attribute_map.items():
				values = getattr(obj, attr_name)
				if values is None:
					continue
				if not isinstance(values, list):
					values = [values]
				attrs[ldap_name] = []
				for value in values:
					if isinstance(value, self.directory.base):
						value = self.directory.models[type(value)].get_dn(value)
					value = encode_attribute(value)
					attrs[ldap_name].append(value)
			attrs['objectClass'] = self.objectclasses
			yield (self.get_dn(obj), attrs)

	def filter_present(self, name):
		if name == 'objectclass':
			return True
		if name not in self.attribute_map:
			return False
		attr = getattr(self.model, self.attribute_map[name])
		if hasattr(attr, 'prop') and isinstance(attr.prop, RelationshipProperty):
			return attr.any()
		return attr.isnot(None)

	def filter_equal(self, name, value):
		if name == 'objectclass':
			return value in self.objectclasses
		if name not in self.attribute_map:
			return False
		attr = getattr(self.model, self.attribute_map[name])
		if isinstance(attr, str):
			value = value.decode()
		elif isinstance(attr, int):
			value = int(value)
		elif isinstance(attr, bytes):
			pass
		elif hasattr(attr, 'prop') and isinstance(attr.prop, RelationshipProperty):
			return attr.any(self.directory.models[attr.prop.argument()].filter_dn(value.decode(), SearchScope.baseObject))
		elif hasattr(attr, 'type') and isinstance(attr.type, String):
			value = value.decode()
		elif hasattr(attr, 'type') and isinstance(attr.type, Integer):
			value = int(value)
		return attr == value

	def _filter_and(self, *subresults):
		return and_(*subresults)

	def _filter_or(self, *subresults):
		return or_(*subresults)

	def _filter_not(self, subresult):
		return ~subresult

	def filter_dn(self, base, scope):
		base = DN(base)
		if scope == SearchScope.baseObject:
			if base[1:] != self.dn_base or len(base[0]) != 1 or base[0][0].attribute != self.rdn_attr:
				return False
			return self.filter_equal(self.rdn_attr, base[0][0].value)
		elif scope == SearchScope.singleLevel:
			return base == self.dn_base
		elif scope == SearchScope.wholeSubtree:
			if self.dn_base.in_subtree_of(base):
				return True
			if base[1:] != self.dn_base or len(base[0]) != 1 or base[0][0].attribute != self.rdn_attr:
				return False
			return self.filter_equal(self.rdn_attr, base[0][0].value)
		else:
			return False

	def get_dn(self, obj):
		rdn_value = getattr(obj, self.attribute_map[self.rdn_attr])
		return str(DN(RDN(**{self.rdn_attr: rdn_value})) + self.dn_base)

class SQLDirectory(BaseDirectory):
	def __init__(self, session, base):
		self.session = session
		self.base = base
		self.models = {}

	def register_model(self, model, attribute_map=None, objectclasses=None, rdn_attr='uid', dn_base=''):
		self.models[model] = SQLModelDirectory(self, model, attribute_map, objectclasses, rdn_attr, dn_base)

	def search(self, baseobj, scope, filter):
		for model, model_directory in self.models.items():
			yield from model_directory.search(baseobj, scope, filter)

